
#include <stdio.h>
/* for getopt */
#include <unistd.h>
/* malloc */
#include <stdlib.h>
/* strlen, memset */
#include <string.h>


#define MAXPAT		65536
#define MAXCHAN		64
#define MASKCHAN	63

int	ptrace, xtrace, vtrace;

short	OrdNum, InsNum, SmpNum, PatNum, Flags, Spec;
unsigned char	*OrdLst;

char	RdMask[MAXCHAN];
char	PreNote[MAXCHAN], PreInst[MAXCHAN], PreVol[MAXCHAN],
	PreEff[MAXCHAN], PreEVal[MAXCHAN];

uint32_t *InsLst, *SmpLst, *PatLst;

void usage(char *myname);
int  it_loadpatt(unsigned char *itbuf, int pattnum, uint16_t *pattlen, uint16_t *pattrows, FILE *in);
int  it_showpatt(unsigned char *itp, int pattnum, uint16_t pattlen, uint16_t pattrows, int start, int maxch, int breakable);
int  it_tochann(int curr, int go_to);
void it_decode_note(char *nna, unsigned char b);
void it_decode_inst(char *nia, unsigned char b);
int  it_decode_vol(char *nva, unsigned char b);
void it_decode_eff(char *nea, unsigned char e, unsigned char v);
int  it_note(char *n, char *i, int ispan, char *v, char *eff);
void printf_bytes(unsigned char *buf, int n, char *msg, char *fmt);
void printf_longs(uint32_t *buf, int n, char *msg, char *fmt);

int main(int argc, char **argv)
{
	FILE	*in;
	int	ch, a;
	uint16_t	tmp1, pattlen, pattrows;
	uint32_t	tmp2;
	unsigned char	*itbuf;
extern	char	*optarg;
extern	int	optind;

	ptrace    = 0;
	xtrace    = 0;
	vtrace    = 0;
	while ((ch = getopt(argc, argv, "xtv")) != -1)
        	switch (ch) {
		case 't':
			ptrace=1;
			vtrace=1; /* what for "PRE" codes? */
			break;
		case 'x':
			xtrace=1;
			break;
		case 'v':
			vtrace=1;
			break;
		default:
			usage(argv[0]);
		}

	itbuf = (char*) malloc(MAXPAT);
	if (itbuf == NULL) return -1;

	if (argc - optind < 1 || (in = fopen(argv[optind],"rb")) == NULL)
		usage(argv[0]);

	/* module identifier */
	fread(itbuf,1,4,in); itbuf[4]=0;
	printf("SongIMPM:[%s]\n",itbuf);

	/* only song name */
	fread(itbuf,1,26,in);
	itbuf[26]=0;
	printf("SongName:[%s]\n", itbuf);
	fread(itbuf,1,2,in); printf_bytes(itbuf, 2, "AlignBytes:", " %02X");

	fread(&OrdNum, 1, 2, in); fread(&InsNum, 1, 2, in);
	fread(&SmpNum, 1, 2, in); fread(&PatNum, 1, 2, in);
	printf("Orders: %d, Instruments: %d, Samples: %d, Patterns: %d\n",
		OrdNum, InsNum, SmpNum, PatNum);

	fread(&tmp1, 1, 2, in); printf("Created w/tracker: %X.%02X\n",
		tmp1 >> 8, tmp1 & 0xFF);

	fread(&tmp1, 1, 2, in); printf("Compatible w/tracker: %X.%02X\n",
		tmp1 >> 8, tmp1 & 0xFF);

	fread(&Flags, 1, 2, in);
	fread(&Spec, 1, 2, in);

	printf("Flags: %04X, Special: %04X\n", Flags, Spec);

	fread(itbuf, 1, 6, in);
	printf("GlobalVolume: %d\nMasterVolume: %d\n"
	"Initial Speed: %d\nInitial Tempo: %d\nPan Separation: %d\n"
	"AlignZero: %d\n",
		itbuf[0], itbuf[1], itbuf[2], itbuf[3], itbuf[4], itbuf[5]);

	fread(&tmp1, 1, 2, in);
	fread(&tmp2, 1, 4, in);
	fread(itbuf, 1, 4, in);

	printf("MessageLength: %d\nMessageAt: %d(0x%X)\n",
		tmp1, (int)tmp2, (int)tmp2);
	printf_bytes(itbuf, 4, "AlignBytes:", " %02X");

	fread(itbuf, 1, 128, in);
	printf_bytes(itbuf+00, 64, "Chan/Pan:", " %3d");
	printf_bytes(itbuf+64, 64, "Chan/Vol:", " %3d");

	OrdLst = (unsigned char*)malloc(OrdNum);
	InsLst = (uint32_t*)malloc(InsNum * sizeof(uint32_t));
	SmpLst = (uint32_t*)malloc(SmpNum * sizeof(uint32_t));
	PatLst = (uint32_t*)malloc(PatNum * sizeof(uint32_t));
	if (!(OrdLst && InsLst && SmpLst && PatLst)) {
		printf("Memory allocation error\n");
		fclose(in);
		return -1;
	}

	fread(OrdLst, 1, OrdNum, in);
	printf_bytes(OrdLst, OrdNum, "Patt.Order:", " %3d");

	fread(InsLst, InsNum, sizeof(uint32_t), in);
	fread(SmpLst, SmpNum, sizeof(uint32_t), in);
	fread(PatLst, PatNum, sizeof(uint32_t), in);

	printf("Object Indexes:\n");
	printf_longs(InsLst, InsNum, "Instrums Offset:", " %-9d (0x%08X)");
	printf_longs(SmpLst, SmpNum, "Samples  Offset:", " %-9d (0x%08X)");
	printf_longs(PatLst, PatNum, "Patterns Offset:", " %-9d (0x%08X)");

	if (ptrace)

	for(a = 0; a < OrdNum; a++) {
	  if (OrdLst[a] == 0xFF) break;
	  if (OrdLst[a] == 0xFE) continue;
	   it_loadpatt(itbuf, OrdLst[a], &pattlen, &pattrows, in);
	ch=it_showpatt(itbuf, OrdLst[a], pattlen, pattrows, ch, -1, 1);
	  OrdLst[a]=255;
	  if ((ch & 0xFF00) == 0x0200) { ch &= 0x00FF; }
	  else if ((ch & 0xFF00) == 0x0300) { a = ch & 0x00FF; ch = 0; }
	  else ch=0;
	}

	else

	for(a = 0; a < PatNum; a++) {
	  it_loadpatt(itbuf, a, &pattlen, &pattrows, in);
	  it_showpatt(itbuf, a, pattlen, pattrows, 0, -1, 0);
	}


	fclose(in);
	return 0;
}

int  it_loadpatt(unsigned char *itbuf, int pattnum, uint16_t *pattlen, uint16_t *pattrows, FILE *in)
{
	  if (PatLst[pattnum] != 0) {
		fseek(in, PatLst[pattnum], SEEK_SET);
		fread(pattlen, 1, sizeof(uint16_t), in);
		fread(pattrows,1, sizeof(uint16_t), in);
		fread(itbuf, 1, 4, in);
	  } else {
		*pattlen=64;
		*pattrows=64;
	  }
	  if(!ptrace)
	  printf("Pattern%03d: (at 0x%X %d bytes, %d rows) ",
		pattnum, PatLst[pattnum], (int)*pattlen, (int)*pattrows);

	  if (PatLst[pattnum] != 0) {
	      if(!ptrace)
		printf_bytes(itbuf, 4, "AlignBytes:", " %3d");
		fread(itbuf, 1, *pattlen, in);
	  } else {
	      if(!ptrace)
		printf("Empty Pattern, virtual\n");
		memset(itbuf, 0, 64);
	  }
}

int it_showpatt(unsigned char *itp, int pattnum,
		uint16_t pattlen, uint16_t pattrows,
		int start, int maxch, int breakable)
{
	unsigned char	*itfin;
	int	ch, is_pan, l, term;
	/* new note action, instr, volume, effect */
	char	nna[4], nia[4], nva[4], nea[4];

	for(ch = 0; ch < MAXCHAN; ch++) RdMask[ch]=0;

	l=-1; term = 0;
	for(itfin = itp+pattlen; itp < itfin;) {
		if (xtrace && l >= 0) printf(" %02x", *itp);
		if (*itp == 0 || l<0) { /* end of row */
			if (l>=0) {
				itp++;
				if (!xtrace)
					{ it_tochann(ch, maxch);printf(" |"); }
				printf("\n");
			}

			ch=0; l++;
			if (term) return term;
			if (l >= pattrows) break;
			if(ptrace)
				printf("%02dp%03d", l, pattnum);
			else
				printf("%02d", l);

			continue;
		}

		/* row read */
		*nna=0; *nia=0; *nva=0; *nea=0;
		if (xtrace)
			ch = ((*itp & 0x7F) - 1) & MASKCHAN; 
		else	/* print empty notes to go to current # */
			ch = it_tochann(ch, ((*itp & 0x7F) - 1) & MASKCHAN); 

		if (*itp & 0x80) { RdMask[ch] = *(++itp);
			if (xtrace) printf(" %02X", *itp); }
		itp++;

		is_pan=0;		
		if (RdMask[ch] & 1) {
			if(xtrace) printf(" %02X", *itp);
			it_decode_note(nna, PreNote[ch]=*itp++);
		}
		if (RdMask[ch] & 2) {
			if(xtrace) printf(" %02X", *itp);
			it_decode_inst(nia, PreInst[ch]=*itp++);
		}
		if (RdMask[ch] & 4) {
			if(xtrace) printf(" %02X", *itp);
			is_pan = it_decode_vol(nva, PreVol[ch]=*itp++);
		}
		if (RdMask[ch] & 8) {
			if(xtrace) printf(" %02X %02X", itp[0], itp[1]);
			it_decode_eff(nea, PreEff[ch]=itp[0], PreEVal[ch]=itp[1]);
			if (breakable && l >= start && (itp[0]==2 || itp[0]==2)) /* && (!term)) - to use first */
				term=((int)itp[1]) | (((int)itp[0]) << 8);
			itp+=2;
		}
		if (vtrace) {
			if (RdMask[ch] & 16) it_decode_note(nna, PreNote[ch]);
			if (RdMask[ch] & 32) it_decode_inst(nia, PreInst[ch]);
			if (RdMask[ch] & 64) is_pan = it_decode_vol(nva, PreVol[ch]);
			if (RdMask[ch] &128) it_decode_eff(nea, PreEff[ch], PreEVal[ch]);
		} else {
			if (RdMask[ch] & 16) sprintf(nna,"PRE");
			if (RdMask[ch] & 32) sprintf(nia,"PR");
			if (RdMask[ch] & 64) sprintf(nva,"PR");
			if (RdMask[ch] &128) sprintf(nea,"PRE");
		}

		if((!xtrace) && l >= start) it_note(nna, nia, is_pan, nva, nea);
		ch++;
		if (ch > maxch) maxch=ch;
	}

	return 0;}

void it_decode_note(char *nna, unsigned char b)
{
static	char* 	notes[12]={ "C-", "C#", "D-", "D#", "E-", "F-", "F#", "G-", "G#", "A-", "A#", "B-" };

	if (b==0xFF) sprintf(nna,"==="); /* noteoff */
	else if (b==0xFE) sprintf(nna, "^^^"); /* note cut */
	else sprintf(nna,"%s%d", notes[b % 12], b / 12);
}

void it_decode_inst(char *nia, unsigned char b)
{
                // Instrument ranges from 1->99

	if (b >=1 && b <=99) sprintf(nia, "%02d",b);
	else sprintf(nia,"??");
}

int it_decode_vol(char *nva, unsigned char b)
{
	if (b <= 64) sprintf(nva,"%02d", b);
	else if (b <=74) sprintf(nva,"E%d", b-65);
	else if (b <=84) sprintf(nva,"F%d", b-75);
	else if (b <=94) sprintf(nva,"D%d", b-85);
	else if (b <=104) sprintf(nva,"U%d", b-95);
	else if (b <=114) sprintf(nva,"L%d", b-105);
	else if (b <=124) sprintf(nva,"R%d", b-115);
	else if (b >= 128 && b <=192)sprintf(nva,"%02d", b-128);
	else sprintf(nva,"??");

	if (b & 0x80) return 1; else return 0;	
}

void it_decode_eff(char *nea, unsigned char e, unsigned char v)
{
	sprintf(nea, "%c%02X", (e == 0)?'.':'A'+e-1, v);
}

int  it_note(char *n, char *i, int ispan, char *v, char *eff)
{
	printf(" | %3s %2s%c%2s %3s", n, i, ispan?'P':' ', v, eff);

	return 0;
}

int  it_tochann(int curr, int go_to)
{
	while(curr < go_to) {
		it_note("", "", 0, "", "");
		curr++;
	}

	return curr;
}


void printf_bytes(unsigned char *buf, int n, char *msg, char *fmt)
{
	int	x, t, l;

	printf("%s",msg);
	l = strlen(msg);
	for(x = 0, t = 0; x < n; x++) {
		if (++t > 16) { printf("\n%*s", l, ""); t=1; }
		printf(fmt, (int)buf[x], (int)buf[x], (int)buf[x], (int)buf[x]);
	}
	printf("\n");
}

void printf_longs(uint32_t *buf, int n, char *msg, char *fmt)
{
	int	x, l;

	printf("%s",msg);
	l = strlen(msg);
	for(x = 0; x < n; x++) {
		if (x) printf("\n%*s", l, "");
		printf(fmt, buf[x], buf[x], buf[x], buf[x]);
	}
	printf("\n");
}


void usage(char *myname)
{
	printf( "Use: %s [-tvx] <module.it>\n"
		"     -t for play trace output but not patterns\n"
		"     -v verbal saved values, do not show PRE\n"
		"     -x for show pattern content in hexadecimal\n", myname);

	exit(1);
}
