
#include <stdio.h>
/* for getopt */
#include <unistd.h>
/* malloc */
#include <stdlib.h>

#define NOTESIZE 4
#define MAXNOTES 4
#define MAXROWS  64
/* minimal object size in file. not system dependent */
#define MINOBJECT 1024

int	modlen;
int	modinsts;
int	modpatts;
int	modinslen[64];
unsigned char playlist[128];
int	notemode;

void usage(char *myname);
int  mod_note(char *buf);

int main(int argc, char **argv)
{
	FILE	*in;
	int	a, l, n, p;
extern	char	*optarg;
extern	int	optind;
	int	ptrace, ch, rowcode;
	long	pattpos, pattsize, mb_size;
	int	notecount, rowcount;
	char	*modbuf, *modp;

	notemode  = 1;
	ptrace    = 0;
	modinsts  = 0;
	modpatts  = 0;
	notecount = MAXNOTES;
	rowcount  = MAXROWS;
	while ((ch = getopt(argc, argv, "st")) != -1)
        	switch (ch) {
		case 's':
			notemode=0;
			break;
		case 't':
			ptrace=1;
			break;
		default:
			usage(argv[0]);
		}

	pattsize  = NOTESIZE * rowcount * notecount;

	mb_size = pattsize;
	if (mb_size < MINOBJECT) mb_size = MINOBJECT;

	modbuf = (char*) malloc(mb_size);
	if (modbuf == NULL) return -1;

	if (argc - optind < 1 || (in = fopen(argv[optind],"rb")) == NULL)
		usage(argv[0]);


	/* only song name */
	fread(modbuf,1,20,in);
	modbuf[20]=0;
	printf("Song:[%s]\n",modbuf);

	/* 31 samples here */
	for(a=1; a < 32; a++) {
		printf("Sample%02d ",a);
		fread(modbuf,1,22,in);
		modbuf[22]=0;
		printf("[%22s] ",modbuf);
		fread(modbuf,1,2,in);
		l = 2 * (((int)(unsigned char)modbuf[1]) | (((int)(unsigned char)modbuf[0]) << 8));
		printf("%6d ",l);
		if (l != 0) modinslen[modinsts++]=l;
		fread(modbuf,1,1,in); if (modbuf[0] & 8) modbuf[0]|=0xf0;
		printf("fine:%2d ", (int) (signed char) modbuf[0]);
		fread(modbuf,1,1,in); modbuf[0]&=127;
		printf("vol:%2d ", (int) modbuf[0]);
		fread(modbuf,1,2,in);
		l = 2 * (((int)(unsigned char)modbuf[1]) | (((int)(unsigned char)modbuf[0]) << 8));
		printf("r.at:%6d ",l);
		fread(modbuf,1,2,in);
		l = 2 * (((int)(unsigned char)modbuf[1]) | (((int)(unsigned char)modbuf[0]) << 8));
		printf("r.sz:%6d\n",l);
	}

	fread(modbuf,1,2,in);
	modlen=(unsigned char)modbuf[0];
	printf("Module length:%d\n", modlen);
	printf("Module repeat at:%d\n", (unsigned char)modbuf[1]);

	fread(playlist,1,128,in);
	for(a=0; a < modlen; a++) {
		if ((a%16) == 0) printf("Playlist: ");
		l = playlist[a];
		if (l > modpatts) modpatts=l;
		printf("%3d ", l);
		if ((a%16) == 15 || (a+1) == modlen) printf("\n");
	}

	printf("Patterns: %d\n", modpatts+1);
	fread(modbuf,1,4,in);
	modbuf[4]=0;
	printf("Signature [%s]\n",modbuf);

	pattpos = ftell(in);

	if (!ptrace) { /* simply dump */

		for(a=0; a <= modpatts; a++) {
		  printf("Pattern%03d:\n",a);
		  fread(modbuf, 1, pattsize, in);

		  for(l = 0, modp=modbuf; l < rowcount; l++) {
			printf("%02d", l+1);
			for(n = notecount; n--;) mod_note(modp+=NOTESIZE);
			printf("\n");
		  }
		}

	} else { /* trace dump, rows only */
		l = 0;
		for(a=0; a < modlen; a++) {
		  p = playlist[a];
		  playlist[a] = 255; /* to prevent repeat */
		  if (p > modpatts) break;

		  fseek(in, pattpos + pattsize * p, SEEK_SET);
		  fread(modbuf, 1, pattsize, in);

		  for(rowcode = 0, modp=modbuf; l < rowcount && !rowcode; l++) {
			printf("%02dp%03d", l+1, p);
			for(n = notecount; n--;)
				if (!rowcode) rowcode=mod_note(modp+=NOTESIZE);
					else mod_note(modp+=NOTESIZE);
			printf("\n");
		  }

		  l = 0;

		  if (rowcode) { /* 0xBXX songpos, 0xDXX break, goto XX row next pattern */
			if (rowcode >= 0xD00 && rowcode <= 0xDFF && (rowcode & 0xFF) < rowcount)
				l = (rowcode & 0xFF);
			if (rowcode >= 0xB00 && rowcode <= 0xBFF && (rowcode & 0xFF) <= modpatts) {
				a = (rowcode & 0xFF);
				a--;
			}
		  }

		}

	}

	fclose(in);
	return 0;
}


int mod_note(char *buf)
{
	char		buf1[6], buf2[4], buf3[4];
	unsigned char	smpno = ((unsigned char)buf[0] & 0xF0);
	unsigned int	period= (((int)(unsigned char)buf[0] & 0x0F) << 8)|((unsigned char)buf[1]);
	unsigned char	eff   = ((unsigned char)buf[2] & 0x0F);
	smpno |= ((unsigned char)buf[2] & 0xF) >> 4;

	if (notemode || period) sprintf(buf1,"%d",period);
		else { buf1[0]='-'; buf1[1]='-'; buf1[2]='-'; buf1[3]=0; }
	if (notemode || smpno)  sprintf(buf2,"%02d",smpno);
		else { buf2[0]='.'; buf2[1]='.'; buf2[2]=0; }
	if (notemode || eff)    sprintf(buf3,"%X",eff);
		else { buf3[0]='.'; buf3[1]=0; }

	printf(" | %-4s %2s %1s %02X", buf1, buf2, buf3, (unsigned char)buf[3]);
	if (eff == 0xB || eff == 0xD) /* break da pattern */
		return (((int)eff) << 8) | (int)(unsigned char)buf[3];
	else
		return 0;
}


void usage(char *myname)
{
	printf( "Use: %s [-s][-t] <module.mod>\n"
		"     -s for simplified output\n"
		"     -t for play trace output but not patterns\n", myname);

	exit(1);
}
